import pyro
from pyro.infer import MCMC, NUTS

from havi.physics.model import (ProbabilisticVariable, ScaterModel, Surface,
                                Wave)


def get_model(
    epsilon: ProbabilisticVariable,
    rms_height: ProbabilisticVariable,
    correlation_longitude: ProbabilisticVariable,
    incident_wave: Wave,
    scatered_wave: Wave,
):
    def model(sigma):
        ep1 = pyro.sample(
            "ep1", pyro.distributions.Uniform(epsilon.range.low, epsilon.range.high)
        )
        s1 = pyro.sample(
            "s1",
            pyro.distributions.Uniform(rms_height.range.low, rms_height.range.high),
        )
        l1 = pyro.sample(
            "l1",
            pyro.distributions.Uniform(
                correlation_longitude.range.low, correlation_longitude.range.high
            ),
        )

        surface = Surface(ep1, s1, l1)
        scater = ScaterModel(surface)
        farlopa = scater.scatering(incident_wave, scatered_wave)
        mu = pyro.deterministic("s0modelMCMC", value=farlopa)
        return pyro.sample("obs", pyro.distributions.Normal(mu, sigma))

    return model


def conditioned_model(model, sigma, y):
    return pyro.poutine.condition(model, data={"obs": y})(sigma)


NUTS_KERNEL = NUTS(conditioned_model, jit_compile=False, ignore_jit_warnings=True)

mcmc = MCMC(
    NUTS_KERNEL,
    num_samples=2000,
    warmup_steps=300,
    num_chains=1,
)



from typing import Any, List, Tuple, Union

import numpy as np
import pandas as pd
import seaborn as sns
import torch

from havi.bayes.inference import get_model, mcmc
from havi.physics.model import EulerAngles, ProbabilisticVariable, Unit, Wave

Tensor = torch.Tensor


class Havi:
    def __init__(
        self,
        incident_angles: List[float],
        scatered_angles: List[float],
        wavelength: float,
    ) -> None:
        self._boudaries_set = False
        theta_i = torch.tensor(incident_angles[0])
        phi_i = torch.tensor(incident_angles[1])
        psi_i = torch.tensor(incident_angles[2])
        theta_s = torch.tensor(scatered_angles[0])
        phi_s = torch.tensor(scatered_angles[1])
        psi_s = torch.tensor(scatered_angles[2])
        landa = torch.tensor(wavelength)
        self.incident_wave = Wave(EulerAngles(theta_i, phi_i, psi_i), landa)
        self.scatered_wave = Wave(EulerAngles(theta_s, phi_s, psi_s), landa)

    def set_boundaries(
        self,
        rms_height: Tuple[float],
        correlation_longitude: Tuple[float],
        dielectric_constant: Tuple[float],
    ) -> None:
        """
        Set boundaries for probabilistic variables

            Parameters:
                rms_height (Tuple[float]): Lower and higher boundaries for rms height values
                correlation_longitude (Tuple[float]): Lower and higher boundaries for correlation long. values
                dielectric_constant (Tuple[float]): Lower and higher boundaries for dielectric constant values

            Returns:
                None
        """

        self.rms_height = ProbabilisticVariable(
            "rms height", Unit.METER, rms_height[0], rms_height[1]
        )
        self.correlation_longitude = ProbabilisticVariable(
            "correlation longitude",
            Unit.METER,
            correlation_longitude[0],
            correlation_longitude[1],
        )
        self.dielectric_constant = ProbabilisticVariable(
            "epsilon", Unit.UNITLESS, dielectric_constant[0], dielectric_constant[1]
        )
        self._boudaries_set = True

    def inference(self, sigma: Tensor, observed: Tensor) -> Union[dict, Any, None]:
        """
        Runs MCMC(Monte Carlo Markov Chain) bayesian inference with NUTS kernel over scatering model

            Parameters:
                sigma (Tensor): error margin for probabilistic model inference
                observed (Tensor): Observed result from scatering measurement to condition bayesian model

            Returns:
                trace (Union[dict, Any, None]): Trace of the inference space of most likely values
        """
        if not self._boudaries_set:
            raise Exception(
                "boundaries for probabilistic variables not set! please call set_boundaries method"
            )
        model = get_model(
            self.dielectric_constant,
            self.rms_height,
            self.correlation_longitude,
            self.incident_wave,
            self.scatered_wave,
        )
        # I am reading this to understand it
        # https://github.com/pyro-ppl/pyro/blob/674af7eab1bb64caf27c9b995913d309229f2c34/pyro/infer/mcmc/api.py#L403
        # observed = torch.Tensor([-12,-7,-8]) #DS2 ### corresponds with Y in conditioned_model
        # ligma = torch.Tensor([0.5]) ## corresponds with sigma in conditioned_model
        mcmc.run(model, sigma, observed)
        mcmc.summary(prob=0.5)
        trace = mcmc.get_samples()
        self.trace = trace
        return trace

    # TODO allow parameter to print or save plots
    def plot(self):
        """
        Plots all of the traced values from MCMC bayesian inference

            Parameters:
                None

            Returns:
                None
        """
        sns.set()
        trace = self.trace
        todos = np.stack((trace["epsilon"], trace["rms_height"], trace["corr_long"])).T
        dfTodos = pd.DataFrame(todos, columns=["eps", "s", "l"])
        # sns.pairplot(dfTodos)
        g = sns.PairGrid(dfTodos)
        g.map_upper(sns.histplot)
        g.map_lower(sns.kdeplot, fill=True)
        g.map_diag(sns.histplot, kde=True)



from dataclasses import dataclass
from enum import Enum
from typing import Dict

from havi.physics.constants import (CONJUGATE, COS, EXP, LOG_10, PI, POW_2,
                                    REAL_PART, SIN, SQRT, ii)


class Unit(Enum):
    UNITLESS = ""
    METER = "m"
    SECONDS = "s"
    METER_PER_SECOND = "m/s"


@dataclass
class Range:
    low: float = 0
    high: float = 1


@dataclass
class Position:
    x: float = 0
    y: float = 0
    z: float = 0


@dataclass
class EulerAngles:
    theta: float = 0
    phi: float = 0
    psi: float = 0


class ProbabilisticVariable:
    def __init__(self, name: str, unit: Unit, low: float, high: float):
        self.name = name
        self.unit = unit
        self.range: Range = Range(low, high)


class Wave:
    def __init__(self, angles: EulerAngles, lambda_: float) -> None:
        self.angles = angles
        self.k = 2 * 3.1416 / lambda_

    def kx(self) -> float:
        return self.k * SIN(self.angles.theta) * COS(self.angles.phi)

    def ky(self) -> float:
        return self.k * SIN(self.angles.theta) * SIN(self.angles.phi)

    def kz(self) -> float:
        return self.k * COS(self.angles.theta)

    def kr(self) -> float:
        return self.k * SIN(self.angles.theta)

    def k_flat(self, epsilon) -> float:
        seno = SIN(self.angles.theta)
        return self.k * SQRT(epsilon - POW_2(seno, 2))


class Floor(Enum):
    ONE = "1"
    TWO = "2"


class WType(Enum):
    GAUSS = "gauss"
    POWER_LAW = "power_law"
    EXPONENTIAL = "exponential"
    CROSS_CORRELATION = "cross"


class Surface:
    def __init__(
        self, epsilon: float, rms_height: float, correlation_longitude: float
    ) -> None:
        self.epsilon = epsilon
        self.rms_height = rms_height
        self.correlation_longitude = correlation_longitude

    def w(self, incident_wave: Wave, scatered_wave: Wave, type: WType) -> float:
        if type == WType.GAUSS:
            return self.w_gauss(incident_wave, scatered_wave)
        elif type == WType.POWER_LAW:
            return self.w_power_law(incident_wave, scatered_wave)
        elif type == WType.EXPONENTIAL:
            return self.w_exp(incident_wave, scatered_wave)
        elif type == WType.CROSS_CORRELATION:
            return self.w_cross(incident_wave, scatered_wave)
        return 0

    def w_gauss(self, incident_wave: Wave, scatered_wave: Wave) -> float:
        """
        Likelyhood of dispersion in the scatered direction given the incident direction

            Parameters:
                incident_wave (Wave): Instance of Wave with incidence values
                scatered_wave (Wave): Instance of Wave with scatered values
                layer (Floor): Enum value matching which ground layer we want to calculate for

            Returns:
                w_gauss (float): Surface potence spectrum dictated by gaussian function
        """
        delta_k_x = scatered_wave.kx() - incident_wave.kx()
        delta_k_y = scatered_wave.ky() - incident_wave.ky()
        rms_height = self.rms_height
        correlation_longitude = self.correlation_longitude
        exponential = EXP(
            -0.25
            * POW_2(correlation_longitude, 2)
            * (POW_2(delta_k_x, 2) + POW_2(delta_k_y, 2))
        )
        return (
            POW_2(rms_height, 2)
            * POW_2(correlation_longitude, 2)
            / (4 * PI)
            * exponential
        )

    def w_exp(self, incident_wave: Wave, scatered_wave: Wave) -> int:
        delta_k_x = scatered_wave.kx() - incident_wave.kx()
        delta_k_y = scatered_wave.ky() - incident_wave.ky()
        rms_height = self.rms_height
        correlation_longitude = self.correlation_longitude
        return (
            POW_2(rms_height, 2)
            * POW_2(correlation_longitude, 2)
            / (
                2
                * PI
                * POW_2(
                    (
                        1
                        + (POW_2(delta_k_x, 2) + POW_2(delta_k_y, 2))
                        * POW_2(correlation_longitude, 2)
                    ),
                    (3 / 2),
                )
            )
        )

    def w_power_law(self, incident_wave: Wave, scatered_wave: Wave) -> int:
        delta_k_x = scatered_wave.kx() - incident_wave.kx()
        delta_k_y = scatered_wave.ky() - incident_wave.ky()
        rms_height = self.rms_height
        correlation_longitude = self.correlation_longitude
        return (
            POW_2(rms_height, 2)
            * POW_2(correlation_longitude, 2)
            / (2 * PI)
            * EXP(
                -SQRT(POW_2(delta_k_x, 2) + POW_2(delta_k_y, 2)) * correlation_longitude
            )
        )

    # def w_cross(self, incident_wave: Wave, scatered_wave: Wave) -> float:
    #     delta_k_x = scatered_wave.kx() - incident_wave.kx()
    #     delta_k_y = scatered_wave.ky() - incident_wave.ky()
    #     s1 = self.layers[Floor.ONE].rms_height
    #     s2 = self.layers[Floor.TWO].rms_height
    #     l1 = self.layers[Floor.ONE].correlation_longitude
    #     l2 = self.layers[Floor.TWO].correlation_longitude
    #     return 0*s1*s2*l1*l2/(4*PI)*EXP(-0.125*(POW_2(l1,2)+POW_2(l2,2))*(POW_2(delta_k_x,2)+POW_2(delta_k_y,2)))


class ScaterModel:
    def __init__(self, surface: Surface) -> None:
        self.surface = surface

    def dispersed_amplitude(self, incident_wave: Wave, scatered_wave: Wave) -> float:
        """this is horrible to read anyway"""
        # calculate once
        eps_1 = self.surface.epsilon
        atm_incidence = self.incidence_atmosphere(incident_wave)
        first_layer_incidence = self.incidence_first_layer(incident_wave)
        incident_flat_wave_eps1 = incident_wave.k_flat(eps_1)
        incident_x_component = incident_wave.kx()
        incident_y_component = incident_wave.ky()
        incident_z_component = incident_wave.kz()
        incident_r_component = incident_wave.kr()
        scatered_flat_wave_eps1 = scatered_wave.k_flat(eps_1)
        scatered_x_component = scatered_wave.kx()
        scatered_y_component = scatered_wave.ky()
        scatered_z_component = scatered_wave.kz()
        scatered_r_component = scatered_wave.kr()
        ikr_sqrd = POW_2(incident_r_component, 2)
        sk_flat_sqrd = POW_2(scatered_flat_wave_eps1, 2)
        exp_layer_scatered_flat_1 = EXP(ii * (2 * scatered_flat_wave_eps1))
        kx_incident_scatered = incident_x_component * scatered_x_component
        ky_incident_scatered = incident_y_component * scatered_y_component
        kx_plus_ky = kx_incident_scatered + ky_incident_scatered
        kz_sqrd = POW_2(incident_z_component, 2)
        kz_plus_kr_sqrd = kz_sqrd + ikr_sqrd
        kxi_minus_kxs = incident_x_component - scatered_x_component
        kyi_minus_kys = incident_y_component - scatered_y_component
        ix_kzkrsqrd_sx = incident_x_component * (kz_plus_kr_sqrd) * scatered_x_component
        scat_x_sqrd = POW_2(scatered_x_component, 2)
        scat_y_sqrd = POW_2(scatered_y_component, 2)
        incident_flat_1_sqrd = POW_2(incident_flat_wave_eps1, 2)
        s_flat_1_minus_z_1 = scatered_flat_wave_eps1 - scatered_z_component * eps_1
        long_term = incident_flat_1_sqrd * (kx_plus_ky) + ikr_sqrd * (
            (kxi_minus_kxs) * scatered_x_component
            + (kyi_minus_kys) * scatered_y_component
        )
        other_long_term = (
            ix_kzkrsqrd_sx
            + incident_y_component * kz_sqrd * scatered_y_component
            - ikr_sqrd * (scat_x_sqrd - ky_incident_scatered + scat_y_sqrd)
        )
        third_long_term = (
            scatered_flat_wave_eps1
            * (kx_plus_ky)
            * (
                (-1 + atm_incidence) * incident_z_component
                + first_layer_incidence * incident_flat_wave_eps1 * SQRT(eps_1)
            )
        )
        exp_sep_incident_flat_1 = EXP(ii * incident_flat_wave_eps1)

        # pieces of the formula
        x = scatered_r_component * (exp_layer_scatered_flat_1) + incident_flat_wave_eps1
        y = third_long_term
        z = sk_flat_sqrd * (kx_plus_ky) * SQRT(eps_1)
        a = (first_layer_incidence) * (long_term)
        b = (1 + atm_incidence) * (other_long_term) * eps_1
        c = -(eps_1) + scatered_flat_wave_eps1
        d = exp_sep_incident_flat_1
        e = third_long_term
        f = first_layer_incidence * sk_flat_sqrd * (kx_plus_ky) * SQRT(eps_1)
        g = first_layer_incidence
        h = long_term
        i = 1 + atm_incidence
        j = other_long_term
        k = eps_1 + scatered_flat_wave_eps1

        q = (
            x * (-y + z - a * SQRT(eps_1) + b) * c
            - d * (-e + f + g * h * SQRT(eps_1) - i * j * eps_1) * k
        )
        r = (
            exp_sep_incident_flat_1
            * incident_r_component
            * (scat_x_sqrd + scat_y_sqrd)
            * (
                EXP(2 * ii * scatered_flat_wave_eps1)
                * (s_flat_1_minus_z_1)
                * (-(eps_1) + scatered_flat_wave_eps1)
                - (scatered_flat_wave_eps1 + scatered_z_component * eps_1)
                * (eps_1 + scatered_flat_wave_eps1)
            )
        )
        return q / r

    def incidence_atmosphere(self, wi: Wave):
        # a0VV
        eps_1 = self.surface.epsilon
        incident_flat_wave_eps1 = wi.k_flat(eps_1)
        incident_z_component = wi.kz()
        incident_flat_1_sqrd = POW_2(incident_flat_wave_eps1, 2)
        q = -ii * incident_flat_wave_eps1 * (incident_z_component) * COS(
            incident_flat_wave_eps1
        ) + (-incident_flat_1_sqrd + incident_z_component) * SIN(
            incident_flat_wave_eps1
        )
        r = ii * incident_flat_wave_eps1 * (incident_z_component) * COS(
            incident_flat_wave_eps1
        ) + (incident_flat_1_sqrd + incident_z_component) * SIN(incident_flat_wave_eps1)
        return q / r

    def incidence_first_layer(self, wi: Wave):
        # b0VV
        eps_1 = self.surface.epsilon
        incident_flat_wave_eps1 = wi.k_flat(eps_1)
        incident_z_component = wi.kz()
        q = -2 * (incident_flat_wave_eps1) * incident_z_component
        r = EXP(2 * ii * incident_flat_wave_eps1) * (incident_flat_wave_eps1) * (
            incident_flat_wave_eps1 - incident_z_component
        ) - (incident_flat_wave_eps1) * (incident_flat_wave_eps1 + incident_z_component)
        return q / r

    def scatering(self, incident_wave: Wave, scatered_wave: Wave) -> float:
        """
        Returns the scatering of a given Channel in decibels.

            Parameters:
                    incident_wave (Wave): Instance of wave with the incidence values
                    incident_wave (Wave): Instance of wave with the scatered values

            Returns:
                    scatering (float): value of scatering in given channel in decibels
        """
        ki = incident_wave.k

        dispersed_amp_1 = self.dispersed_amplitude(incident_wave, scatered_wave)

        w_1 = self.surface.w(incident_wave, scatered_wave, WType.GAUSS)

        aux = (
            4
            * PI
            * POW_2(ki, 2)
            * POW_2(COS(scatered_wave.angles.theta), 2)
            * (POW_2(abs(dispersed_amp_1), 2) * w_1)
        )

        return 10 * LOG_10(aux)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['havi', 'havi.bayes', 'havi.physics']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.21.5,<2.0.0',
 'pandas>=1.3.5,<2.0.0',
 'pyro-ppl>=1.8.0,<2.0.0',
 'seaborn>=0.11.2,<0.12.0',
 'torch>=1.10.1,<2.0.0']

setup_kwargs = {
    'name': 'havi',
    'version': '0.1.1',
    'description': 'perform bayesian inference over physical models',
    'long_description': '[![Version](https://img.shields.io/pypi/v/havi)](https://pypi.org/project/havi)\n# Havi\n[stands for Odin](https://en.wikipedia.org/wiki/H%C3%A1r_and_H%C3%A1rr)\n\n\nscatering model and bayesian inference\n\n---\n**Note from author**\n\nIf you use our library on a research or paper please give us a citation\n\n- [PyPi](https://pypi.org/project/havi/)\n\n- [Github](https://github.com/bjornaer/havi)\n\n---\n\n### Installation\n\n    pip install havi\n\n### Usage\n\n```python\nimport havi as h\n\n# scatering angles\nincident_angles = [theta, phi, psi] # incidence angles - floats\nscatered_angles = [theta, phi, psi] # scatering angles - floats\nwavelength = 0.5\n\nhavi = h.Havi(incident_angles, scatered_angles, wavelength)\n\n# surface boudaries for random variables\nrms_height = (0.1, 0.8) # lower and higher boundaries\ncorrelation_longitude = (0.1, 0.8) # lower and higher boundaries\ndielectric_constant = (0.1, 0.8) # lower and higher boundaries\n\nhavi.set_boundaries(rms_height, correlation_longitude, dielectric_constant)\n\nsigma = h.Tensor([0.5])\nobserved_data = h.Tensor([-12,-7,-8])\n\ntrace = havi.inference(sigma, observed) # you can plot the trace however you want OR\nhavi.plot()\n```\n\n### API\n\nThis package exports:\n\nA pytorch Tensor type\n\n    havi.Tensor\n\nOur inference model abstraction, holds data of incident wave, scatered wave and their wave length:\n        \n    havi.Havi(incident_angles: List[float], scatered_angles: List[float], wavelength: float)\n\nSet boundaries for probabilistic variables\n\n    Havi.set_boundaries(rms_height: Tuple[float], correlation_longitude: Tuple[float], dielectric_constant: Tuple[float]) -> None\n\nRuns MCMC(Monte Carlo Markov Chain) bayesian inference with NUTS kernel over scatering model\n\n    Havi.inference(sigma: Tensor, observed: Tensor) -> trace: dict\n\nPlots all of the traced values from MCMC bayesian inference\n\n    Havi.plot() -> None\n\n',
    'author': 'bjornaer',
    'author_email': 'maxemijo@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



import numpy as np
import torch


def is_numpy(x):
    if type(x).__module__ == "numpy":
        return True
    return False


ii = torch.tensor(1j, dtype=torch.cfloat)
# abstract math functions if we need to stop using numpy
PI = np.pi
EXP = lambda x: np.exp(x) if is_numpy(x) else torch.exp(x)
SQRT = lambda x: np.sqrt(x) if is_numpy(x) else torch.sqrt(x)
SIN = lambda x: np.sin(x) if is_numpy(x) else torch.sin(x)
COS = lambda x: np.cos(x) if is_numpy(x) else torch.cos(x)
REAL_PART = lambda x: np.real(x) if is_numpy(x) else torch.real(x)
CONJUGATE = lambda x: np.conj(x) if is_numpy(x) else torch.conj(x)
LOG_10 = lambda x: np.log10(x) if is_numpy(x) else torch.log10(x)
POW_2 = torch.pow
